"Python3 document categorization library, n-gram coverage method."

##  TAVERNS (Textual Affinity Values Employing Repeated N-gram Sequences) :
##
##  Python3 data & routines to supply to toccata test harness s/w :
##  Copyright 2017, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 20/01/2017
##  last revision : 26/01/2017


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys
import  random


if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
    

from globdefs import *
from glib import  add2dic, numerize
from tocclib import  bastats
from tocclib import  corsums, corrval
from tocclib import  midvalue, ranx
from tocclib import  fourfold, tetracor


##  global vars :
Showoff = 0  ##  >0 for debugging o/p
Gold = 5.0**0.5*0.5+0.5  ##  good approximation to golden number
Nulltok = ".d0t."
Nulltox = "D0T.D0T$"
Maxtoks = 48 * 1024  ##  slow with large files, so truncate


##  functions first, then class(es) :

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  TOPTOKS :
def toptoks (freqtab,h=144):
    "returns set of high-frequency tokens from frequency dict."

    v = len(freqtab)  ##  vocsize
    sortlist = []
    for k in freqtab:
        f = freqtab[k]
        if f > 1:  ## always skip hapaxes
            sortlist.append((f,k))
    sortlist.sort(reverse=True)

    return  sortlist[:h]


##  ADDSNIPS :
def addsnips (d,snipsize):
    "adds snippet data to text d if not already present."

    d.snippets = 0
    d.snipsets = []  ##  list of sets
    minsnip = int(round(snipsize / Gold))
    for p in range(0,d.size,snipsize):
        slab = d.toklist[p:p+snipsize]
        if len(slab) >= minsnip or p==0:
            d.snipsets.append(set(slab))
            d.snippets += 1

    return  ##  side-effect on d.


##  G2SCORE :
def g2score (n,frex,nulrates):
    "quick & dirty g-squared score from frex & base rates."

    ##  n should be sum(frex).
    assert len(frex) == len(nulrates)
    g2 = 0.0
    for j,f in enumerate(frex):
        if frex[j] > 0:
            e = n * nulrates[j]
            gx = frex[j] * math.log(frex[j]/e)
            ##  e must never be zero !
            g2 += gx

    return  g2


##  TRIMTOX :
def trimtox (toklist,trimmode,tops):
    "trims less common words from toklist."
    
    if trimmode == 0:  ##  just use token list
        trimtoks = toklist
    elif trimmode == 1:  ##  omit less frequent tokens
        trimtoks = [t for t in toklist if t in tops]
    else:  ##  mode 2 : replace less frequent tokens by nulls
        dt = []
        for t in toklist:
            if t in tops:
                dt.append(t)
            else:
                dt.append(Nulltok)
        trimtoks = dt
        if trimmode == 3:  ##  mode 3 : distinguish singles from multiples
            trimtoks = ['']
            nt = len(dt)
            for t in dt:
                if trimtoks[-1] == Nulltox and t == Nulltok:
                    pass  ##  skip null reps
                elif trimtoks[-1] == Nulltok and t == Nulltok:
                    trimtoks[-1] = Nulltox
                    ##  change to Nulltox.
                else:
                    trimtoks.append(t)
            del trimtoks[0]  ##  drop temporary prefix
        if trimmode == 4:  ##  mode 4 : Nulltox for 1..n Nulltoks
            trimtoks = ['']
            nt = len(dt)
            for t in dt:
                if t == Nulltok and trimtoks[-1] == Nulltox:
                    pass  ##  don't repeat
                elif t == Nulltok:
                    trimtoks.append(Nulltox)
                    ##  change to Nulltox.
                else:
                    trimtoks.append(t)
            del trimtoks[0]  ##  drop temporary prefix
            
    return  trimtoks


##  FREQLEN :
def freqlen (gtab):
    "orders gtab by frequency then length."

    sortlist = []
    for g in gtab:
        f = gtab[g]
        nc = len(' '.join(g))
        if f > 2:  ##  hapaxes & dislegomena dropped
            sortlist.append((f,len(g),nc,g))

    sortlist.sort(reverse=True)
    return  sortlist
    ##  format : (freq,toksize,charsize,gram).


##  GRAMSEEK :
def gramseek (doclist,catname,miniglen,maxiglen):
    "seeks n-grams in a given category."

    gramtab = {}
    for d in doclist:
        if d.outcome != catname:
            continue  ##  skip other classes of document
        trimlen = len(d.trimtoks)
        for i in range(trimlen):
            slab = d.trimtoks[i:i+maxiglen]
            ##  print(i,slab)
            for j in range(miniglen,len(slab)+1):
                ##  print(slab[0:j])
                gram = tuple(slab[0:j])
                if gram in gramtab:
                    gramtab[gram] += 1
                else:
                    gramtab[gram] = 1

    return  gramtab


##  GOUT :
def gout (toplist,miniglen,maxiglen,topgrams,fo):
    "puts out sorted toplist of n-grams by size."

    for n in range(maxiglen,miniglen-1,-1):
        ##  step down by # of tokens.
        counter = 0
        for t in toplist:
            freq,size,nc,g = t
            if size != n:
                continue
            ##  doesn't omit subgrams @ present
            counter += 1
            print(counter,t,file=fo)
            if counter >= topgrams:
                break  ## enough done at this size
            
    return


##  COVERED :
def covered (d,markmode=0):
    "finds coverage by n-grams in 1 document."

    n = len(d.marx)
    ##  assert n == d.size  ##  no longer true
    ch = th = 0
    for p in range(n):
        if d.marx[p] > 0:  ##  covered by frequent n-grams
            if markmode == 0:
                th += 1  ##  binary: absence/presence
            elif markmode == 1:
                th += d.marx[p]  ##  weighted as longest covering n-gram
            else:  ##  mode 2
                th += math.sqrt(d.marx[p])  ##  attenuated ditto.
            ch += (len(d.trimtoks[p])+1)  ##  always count (trailing) space

    return  ch,th


##  GRAMMARX :
def grammarx (gramlist,miniglen,maxiglen,toks):
    "marks locations in d covered by n-grams of type catname."

    ntox = len(toks)  ##  may have been trimmed
    marx = [0] * ntox
    for item in gramlist:
        g = list(item[-1])  ##  n-gram itself @end, forced 2b a list
        glen = len(g)  ##  can't be zero
        if glen < miniglen:
            continue  ##  too short
        locs = findlocs(toks,g[0])  ##  find locs of initial token
        for p in locs:
            slab = toks[p:p+glen]
            if slab == g:
                for j in range(p,p+glen):
                    ##  marx[j] += 1  ##  hits at this token position
                    marx[j] = max(marx[j],glen)  ##  longest covering n-gram

    return  marx  ##  2b held with doclist element


##  FINDLOCS :
def findlocs (toklist,item):
    "returns locations of item in token list."

    locs = []
    p = 0
    nt = len(toklist)
    while p < nt:
        try:
            p = toklist.index(item,p)
            locs.append(p)
            p += 1
        except ValueError as xx:
            break

    return  locs




####  DOCADAT :
class Docadat (object):
    ##  collection of data+methods to do document classification :
    ##  assumes access to 2 Sack() objects as args, doclist & paradat.
    ##  doclist attributes expected include:
    ##    classnum, freqtab, outcome, size, text, textlen, toklist,
    ##  paradat attributes expected include:
    ##    catlist, cats, paraline, testmeta, trainmet,

    def __init__ (self,paradat,doclist):
        "initialization method for Docadat objects."

        self.classes = paradat.cats
        self.catlist = paradat.catlist
        self.id = paradat.trainmet
        if hasattr(paradat,"testmeta"):
            self.id = self.id + '_' + paradat.testmeta
        self.alldocs = len(doclist)  ##  number of training-corpus documents

        return

    def loadpars (self,paradat,sep1=' ',sep2='='):
        "loads extra docalib parameters if any, provided in paraline."

        parline = paradat.paraline.lower().strip()
        self.pars = Sack(parline)
        parbits = parline.split(sep1)  ##  item separator defaults to space
        for pb in parbits:
            bits = pb.split(sep2)  ##  internal separator (e.g. "gramsize=4")
            if len(bits) == 2:
                v = numerize(bits[1])
                if type(v) == type(9) or type(v) == type(0.5):  ##  numeric
                    setattr(self.pars,bits[0],v)
                else:  ##  nonnumeric
                    setattr(self.pars,bits[0],bits[1])
                ##  no limits imposed !
        ##  check for suggested tops2use :
        tv = 0
        if hasattr(self.pars,"topvocs") and isinstance(self.pars.topvocs,int):
            self.topvocs = getattr(self.pars,"topvocs")
        else:
            self.topvocs = tv
        ##  tops2use won't use this if < 1 or > 256.
        tm = 0
        if hasattr(self.pars,"trimmode") and isinstance(self.pars.trimmode,int):
            self.trimmode = getattr(self.pars,"trimmode")
        else:
            self.trimmode = tm  ##  default
        self.trimmode = min(4,max(0,self.trimmode))  ##  enforce limits
        ##  0 use toklist; 1 remove 'rare' toks; 2 replace 'rare' toks by nils;
        ##  3 distinguish single from multiple nils.
        ##  coverage mode :
        if not hasattr(self.pars,"markmode"):
            self.markmode = 0  ##  default to absence/presence
        elif isinstance(self.pars.markmode,int):
            self.markmode = self.pars.markmode
        else:
            self.markmode = 0  ##  default
        self.markmode = min(2,max(0,self.markmode))  ##  enforce limits
        ##  0 covered 1/uncovered 0; 1 wt by size of longest covering n-gram;
        ##  2 weighted by sqrt of size of longest covering n-gram.
        ##  print("markmode",self.markmode)
        ss = 144  ##  default snipsize
        if hasattr(self.pars,"snipsize") and isinstance(self.pars.snipsize,int):
            self.pars.snipsize = numerize(getattr(self.pars,"snipsize",ss))
        else:
            self.pars.snipsize = ss  ##  may still be altered if unrealistic
        self.miniglen = 2
        if hasattr(self.pars,"miniglen") and isinstance(self.pars.miniglen,int):
            self.miniglen = self.pars.miniglen
        self.miniglen = min(5,max(1,self.miniglen))
        self.maxiglen = 4
        if hasattr(self.pars,"maxiglen") and isinstance(self.pars.maxiglen,int):
            self.maxiglen = self.pars.maxiglen
            if self.maxiglen > 9:
                self.maxiglen = 9
        self.maxiglen = max(self.miniglen,self.maxiglen)
        
        return  ##  doesn't crash if paradat.paraline empty

    def modprep (self,paradat):
        "does once-only overall tasks."

        ##  check that snipsize is reasonable :
        ns = paradat.tottoks / self.pars.snipsize
        if ns < 64 or ns > 8192:
            self.pars.snipsize = int(paradat.tottoks ** 0.5)
            print("snipsize altered to",self.pars.snipsize)
        else:
            print("snipsize =",self.pars.snipsize)
        ##  attach snippets to docs :
        totsnips = 0
        minsnip = int(round(self.pars.snipsize / Gold))
        for d in paradat.doclist:  ##  deals with entire sample
            d.toklist = d.toklist[:Maxtoks]  ##  enforce size limit
            d.snippets = 0
            d.snipsets = []  ##  list of sets
            for p in range(0,d.size,self.pars.snipsize):
                slab = d.toklist[p:p+self.pars.snipsize]
                if len(slab) >= minsnip or p==0:
                    d.snipsets.append(set(slab))
                    d.snippets += 1
                    totsnips += 1
        print("Number of snippets =",totsnips)
        paradat.totsnips = totsnips  ##  does belong to whole corpus
        ##  now find high-freq tokens :
        tokfrex = {}
        for d in paradat.doclist:
            for snip in d.snipsets:
                for t in snip:
                    if t in tokfrex:
                        tokfrex[t] += 1
                    else:
                        tokfrex[t] = 1
            d.snippets = len(d.snipsets)  ##  record number of snippets in d
        paradat.highfrex = toptoks(tokfrex,1024)
        ##  compile overall tables once only, at start :
        self.fullfrex = {}  ##  overall training-corpus frequency table
        ##  compile overall freq table :
        for d in paradat.doclist:
            add2dic(self.fullfrex,d.freqtab)
        self.docs = len(paradat.doclist)
        self.vocsize = len(self.fullfrex)  ##  vocabulary size
        ##  used to decide number of common terms to keep.
        self.tops2use = int(math.sqrt(self.vocsize))  ##  desired number of vocs
        if self.topvocs >= 2 and self.topvocs < 256:  ##  sensible
            self.tops2use = self.topvocs
        if self.tops2use > 256:
            self.tops2use = 256
        ##  otherwise stick with heuristic/default.
        self.goldtops = round(self.tops2use / Gold)
        ##  self.goldtops = self.tops2use
        self.toplist = paradat.highfrex[:self.tops2use]
        self.toptoks = [x[1] for x in self.toplist]  ##  without frex
        self.tops = set(self.toptoks)
        ##  high-frequency vocs, i.e. features.
        ##  apply preprocessing of less common tokens to toklists:
        self.trimdocs(paradat,paradat.doclist)
        ##  each document now has .trimtoks field
        
        return
    
    def trimdocs (self,paradat,doclist):
        "appends trimmed toklist to each document."

        for d in doclist:
            d.trimtoks = trimtox(d.toklist,self.trimmode,self.tops)
            ##  too slow 2use very long texts entire.
            
        return
    
    def makemods (self,paradat,doclist):
        "builds classification models for each category."
        ##  doclist will be training data.

        self.modlist = []  ##  list of category models
        self.modeltab = {}  ##  dictionary of models, if lookup by name needed
        self.name2num = {}

        ##  make models :
        cn = 0
        for c in self.catlist:  ##  relies on consistent order of catlist
            m = self.makemod(c,cn,paradat,doclist)
            self.modlist.append(m)
            self.modeltab[c] = m  ##  also point from dict keyed by catname
            self.name2num[c] = cn
            cn += 1  ##  keep track of position in model list        
        self.mods = len(self.modlist)
        assert  self.mods == self.classes

        return

    def makemod (self,c,cn,paradat,doclist):
        "builds model info for a single category."

        m = Sack(c)  ##  gathered into a Sack() object
        m.tottoks = 0
        m.catname = c
        m.classnum = cn
        m.gramtab = gramseek(doclist,c,self.miniglen,self.maxiglen)
        m.gramlist = freqlen(m.gramtab)
        ##  only keep frequent items at each length :
        m.toplist = []
        for glen in range(self.miniglen,self.maxiglen+1):
            ng = 0
            for gram in m.gramlist:
                g = gram[-1]  ##  n-gram @end
                if len(g) == glen:
                    m.toplist.append(gram)
                    ng += 1
                if ng >= self.goldtops:
                    break  ##  enough at this size
        show = False  ##  no dump
        for d in doclist:
            if d.outcome == c:
                ##  safer for this library not2 cause side-effects in caller !
                m.tottoks += d.size
        ##  print(m.catname,len(m.gramtab),len(m.gramlist),len(m.toplist))
        print(m.catname,len(m.gramtab),len(m.gramlist),len(m.toplist),
              file=paradat.fo)
        ##  print(m.toplist,file=paradat.fo)
        ##  .gramtab now superfluous ??
        
        return  m

    def showmod (self,cn,fo=sys.stdout):
        "shows info related to model of category number cn."

        m = self.modlist[cn]
        assert cn == m.classnum
        print("category:",m.catname,'(',cn,')',file=fo)
        print("tokens =",m.tottoks,file=fo)
        gout(m.toplist,self.miniglen,self.maxiglen,self.tops2use,fo)

        print('',file=fo)

        return

    def showmods (self,paradat,fo=sys.stdout):
        "shows models of all categories."

        print("classes =",self.classes,file=fo)
        print(paradat.catlist,file=fo)
        print("docs =",paradat.docs,file=fo)
        print("tokens =",paradat.tottoks,file=fo)
        print("vocsize =",self.vocsize,file=fo)
        print("topterms =",self.tops2use,file=fo)
        print("trimmode =",self.trimmode,file=fo)
        print("markmode =",self.markmode,file=fo)
        print("subsizes :",self.miniglen,self.maxiglen,file=fo)
        print("toplist :",self.toplist,file=fo)
        print("toptoks :",self.toptoks,file=fo)
        print(self.tops,file=fo)
        print('',file=fo)

        for cn in range(self.classes):
            self.showmod(cn,fo)
            
        return
        ##  doesn't open or close o/p file; caller must do that.

    def matchmod (self,cn,thisdoc,paradat):
        "matches model for category number cn to given document."

        m = self.modlist[cn]
        if not hasattr(thisdoc,"trimtoks"):  ##  presume test-data item
            thisdoc.trimtoks = trimtox(thisdoc.toklist[:Maxtoks],self.trimmode,
                                   self.tops)
        thisdoc.marx = grammarx(m.toplist,self.miniglen,self.maxiglen,
                                thisdoc.trimtoks)
        ch,th = covered(thisdoc,self.markmode)

        if thisdoc.dnum == 1:
            print(thisdoc.trimtoks[:99],file=paradat.fo)
            print(thisdoc.marx[:99],file=paradat.fo)
            print(len(thisdoc.trimtoks),len(thisdoc.marx),th,file=paradat.fo)
            
        return  th/len(thisdoc.trimtoks)  ##  matching score, higher implies stronger match

    
    def modsims (self,thisdoc,paradat):
        "returns association score of each model to this doc."

        catvals = [0.0] * paradat.cats
        for cpos,c in enumerate(paradat.catlist):
            catvals[cpos] = self.matchmod(cpos,thisdoc,paradat)

        return  catvals
        ##  no scaling to quasi-probs at present.
        ##  return  [1.0/(1e-7+catvals[cpos]) for cpos in range(self.classes)]
        ##  inversion of distances used, so more means better.
