"Python3 document categorization library, vote method (with tokens)."

##  DOCALIB :
##
##  Python3 data & routines to supply to toccata test harness s/w :
##  Copyright 2013, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 16/01/2013
##  last revision : 24/04/2017


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys

if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
    

from globdefs import *
from glib import  add2dic, numerize
from tocclib import  corsums, corrval
from tocclib import  ranx


##  global vars :
Showoff = 0  ##  for debugging o/p
Gold = 5.0**0.5*0.5+0.5  ##  good approximation to golden number (redundant?)


##  functions first, then class(es) :


##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  TOPTOKS :
def toptoks (freqtab,h=144):
    "returns set of high-frequency tokens from frequency dict."

    v = len(freqtab)  ##  vocsize
    sortlist = []
    for k in freqtab:
        f = freqtab[k]
        if f > 1:  ## always skip hapaxes
            sortlist.append((f,k))
    sortlist.sort(reverse=True)

    return  sortlist[:h]



####  DOCADAT :
class Docadat (object):
    ##  collection of data+methods to do document classification :
    ##  assumes access to 2 Sack() objects as args, doclist & paradat.
    ##  doclist attributes expected include: classnum, freqtab, outcome, size,
    ##    text, textlen, toklist,
    ##  paradat attributes expected include:
    ##    catlist, cats, paraline, testmeta, trainmet,

    def __init__ (self,paradat,doclist):
        "initialization method for Docadat objects."

        self.classes = paradat.cats
        self.catlist = paradat.catlist
        self.id = paradat.trainmet
        if hasattr(paradat,"testmeta"):
            self.id = self.id + '_' + paradat.testmeta
        self.alldocs = len(doclist)  ##  number of training-corpus documents

        return

    def loadpars (self,paradat,sep1=' ',sep2='='):
        "loads extra docalib parameters if any, provided in paraline."

        parline = paradat.paraline.lower().strip()
        self.pars = Sack(parline)
        parbits = parline.split(sep1)  ##  item separator defaults to space
        for pb in parbits:
            bits = pb.split(sep2)  ##  internal separator (e.g. "gramsize=4")
            if len(bits) == 2:
                v = numerize(bits[1])
                if type(v) == type(9) or type(v) == type(0.5):  ##  numeric
                    setattr(self.pars,bits[0],v)
                else:  ##  nonnumeric
                    setattr(self.pars,bits[0],bits[1])
                ##  no limits imposed !
        ##  determine whether to attenuate frequencies :
        if not hasattr(self.pars,"rootfrex"):
            self.pars.rootfrex = 1  ##  default
        if not hasattr(self.pars,"flatfrex"):
            self.pars.flatfrex = 0  ##  default
        return  ##  doesn't crash if paradat.paraline empty
    
    def makemods (self,paradat,doclist):
        "builds classification models for each category."

        self.modlist = []  ##  list of category models
        self.modeltab = {}  ##  dictionary of models, if lookup by name needed
        self.name2num = {}
        self.fullfrex = {}  ##  overall training-corpus frequency table
        self.tottoks = 0
        cn = 0
        for c in self.catlist:  ##  relies on consistent order of catlist
            m = self.makemod(c,cn,paradat,doclist)
            self.modlist.append(m)
            self.modeltab[c] = m  ##  also point from dict keyed by catname
            self.name2num[c] = cn
            add2dic(self.fullfrex,m.freqtab)
            self.tottoks += m.tottoks  ##  total of all (word-)tokens
            cn += 1  ##  keep track of position in model list
        
        self.mods = len(self.modlist)
        self.docs = len(doclist)
        assert  self.mods == self.classes
        self.vocsize = len(self.fullfrex)  ##  overall vocabulary size
        if Showoff > 0:
            print(self.tottoks,self.vocsize,'\n',file=paradat.logfile)
        
        return

    def makemod (self,c,cn,paradat,doclist):
        "builds model info for a single category."

        m = Sack(c)  ##  gathered into a Sack() object
        m.freqtab = {}
        m.tottoks = 0
        m.catname = c
        m.classnum = cn
        for d in doclist:
            if d.outcome == c:
                add2dic(m.freqtab,d.freqtab)
                ##  freqtabs expected 2b computed by calling program.
                ##  safer for this library not2 cause side-effects in caller.
                m.tottoks += d.size
        m.vocsize = len(m.freqtab)
        ##  each class also has a high-freq list.
                                   
        return  m

    def showmod (self,cn,fo=sys.stdout):
        "shows info related to model of category cn."

        m = self.modlist[cn]
        assert cn == m.classnum
        print("category:",m.catname,'(',cn,')',file=fo)
        print("tokens =",m.tottoks,file=fo)
        print("vocsize =",len(m.freqtab),file=fo)
        fulltoks = self.tottoks
        termlist = []
        for t in m.freqtab:
            f0 = self.fullfrex.get(t,0)
            f1 = m.freqtab.get(t)
            r0 = f0 / fulltoks
            r1 = f1 / m.tottoks
            termlist.append((100.0*(r1-r0),f1,t))
        j = 1
        for r,f,t in sorted(termlist,reverse=True):
            if r <= 0.000005:  break
            if f <= 2:
                continue  ##  skip haps & dislegs
            lout = "{0:4d} {1:8.4f} {2:5d} {3:20s}".format(j,r,f,t)
            print(lout,file=fo)
            j += 1

        print('',file=fo)
        return

    def showmods (self,paradat,fo=sys.stdout):
        "shows models of all categories."

        print("classes =",self.classes,file=fo)
        print(paradat.catlist,file=fo)
        print("docs =",paradat.docs,file=fo)
        print("tokens =",paradat.tottoks,file=fo)
        ##  assert paradat.tottoks == sum(self.fullfrex.values())
        print("vocsize =",self.vocsize,file=fo)
        fo.write(NL)
        for cn in range(self.classes):
            self.showmod(cn,fo)

        return
        ##  doesn't open or close o/p file; caller must do that.

    def matchmod (self,cn,thisdoc,paradat):
        "matches model for category number cn to given document."

        votesum = 0.0
        doctoks = thisdoc.size
        assert doctoks == len(thisdoc.toklist)
        divisor = 0
        fulltoks = self.tottoks
        modtoks = self.modlist[cn].tottoks
        errmess = False
        if Showoff > 0:
            print(doctoks,modtoks,fulltoks,thisdoc.name,file=paradat.logfile)
        for w in thisdoc.freqtab:
            docfreq = thisdoc.freqtab[w]
            modfreq = self.modlist[cn].freqtab.get(w,0)
            fullfreq = self.fullfrex.get(w,0)
            ##  if modfreq < 2:  continue
            f0 = fullfreq
            n0 = fulltoks
            f1 = modfreq ; n1 = modtoks
            if f1 < 0 or n1 < 1 or f0 < 0 or n0 < 1:
                if not errmess:
                    print(f0,n0,f1,n1)
                    print("Trouble with counts in matchmod() !!")
                    errmess = True
                n1 = max(1,n1) ; n0 = max(1,n0)
            r0 = f0 / n0 ; r1 = f1 / n1  ##  rates
            ##  roar = (1.0-r0) / (1.0-r1)
            ##  vote = math.tanh(r1-r0)
            ##  vote = roar
            vote = 0.5 + 0.5*(r1>r0) - 0.5*(r0>r1)
            rf = docfreq
            if self.pars.rootfrex > 0:
                rf = rf ** 0.5
            if self.pars.flatfrex > 0:  ##  takes precedence
                rf = 1  ##  just count each token equally !
            votesum += (vote * rf)  ##  used 2b math.log(.)
            divisor += rf

        if self.pars.rootfrex == 0:  ##  no longer default
            ff = getattr(self.pars,"flatfrex",0)
            if ff > 0:
                assert divisor == len(thisdoc.freqtab)
            else:
                assert divisor == doctoks
            ##  sanity checks.
        return  votesum/divisor  ##  matching score, higher implies stronger match

    def modsims (self,thisdoc,paradat):
        "returns matching scores of this text againts each class model."

        catvals = [0.0] * paradat.cats
        for cpos,c in enumerate(paradat.catlist):
            catvals[cpos] = self.matchmod(cpos,thisdoc,paradat)

        return  catvals
        ##  no auto scaling at present.
        ##  also no inversion if distances used, thus here more is better.
