"Python3 test harness for toccata software."

##  TOCCATA :
##
##  Text-Oriented Computational Classifier Applicable To Authorship :
##  Copyright 2013, R.S. Forsyth.
##  first version : 15/01/2013
##  last revision : 22/04/2017


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
    This copyrighted program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math, os, string, sys
import  random, time
##  import  copy


if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
##  in case executed from somewhere else.
    

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow, paracopy
from glib import  showobj, paradump
from glib import  openfile
from glib import  punktran
from glib import  metachex, metagrab
from glib import  niceresp, readmeta
from glib import  getdocs
from glib import  vecline
from glib import  pybyebye
from tocclib import  bastats
from tocclib import  corsums, corrval
from tocclib import  midvalue
from tocclib import  sckappa2


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("filextn",''),
    ("targvar","doctype"),  ##  name of target column with outcome values
    ("targcol",0,0),  ##  column 2use as outcome
    ("jobname","toccata"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("wordonly",0,0,1),  ##  whether to look only at tokens with alfnum initial
    ("randseed",1789,1,999999),  ##  random seed, for replicability
    ("fulldump",0,0,1),  ##  whether to dump or just show parameter values
    ("libname","docalib_topvocs"),  ##  name of library with classifier s/w
    ("paraline",''),  ##  free-form parameters to pass to library s/w
    ("metafile",''),  ##  training corpus
    ("trainmet",''),  ##  ditto (takes priority over metafile)
    ("testmeta",''),  ##  test corpus (optional)
    ("dumpfile",''),  ##  machine-readable results file
    ("postcode",0,0,1),  ##  0 -> write posthoc file anew; 1 -> append to it
    ("listfile",''),  ##  human-readable results file
    ("modsfile",''),  ##  file for o/p of model info
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"toccata.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "9.2"
Gold = 5.0**0.5*0.5+0.5
Mindocs = 4
Dumptail = "_dump"
Listtail = "_list"
Logtail = "_tocpars"
Modtail = "_mods"
Posttail = "_posthoc"
Longname = 28  ##  longest class name; longer will be truncated
Longfile = 26  ##  longest file name for o/p; longer will be truncated
Maxtoks = 64 * 1024  ##  will only seek tokens in front chunk of text file
Mintries = 255  ##  minimum # of subsampling trials
Plusmark = '+'
Negmark = '-'
Oddmark = '?'


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()

    random.seed(paradat.randseed)  ##  ensure repeatability
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.path.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
    if not os.path.isfile(paradat.trainmet):  ##  no explicit training metafile
        if os.path.isfile(paradat.metafile):  ##  fall back on metafile
            paradat.trainmet = paradat.metafile
            print("?? Possible problem: no training file specified.")
            print("Using metafile",paradat.metafile,"instead.")
        else:
            paradat.trainmet = ''  ##  ensure early ending
    ##  parashow(paradat)
    tinf = Tokstuff(paraname)
    paradat.tokinfo = tinf  ##  tokenization info passed thru
    paradat.punxtab = punktran()  ##  uses default equivalents

    ##  read meta-data file of training corpus :
    metaname = getattr(paradat,"trainmet",'')
    if not metaname:
        print("Can't proceed without trainmet, i.e. training metafile!")
        print("Program halts.")
        parashow(paradat)
        return  -1

    outcomes = metagrab(metaname,paradat)
    paradat.outcomes = outcomes
    doclist = getdocs(paradat,paradat.filelist,paradat.outcomes,Maxtoks)
    ##  size limit enforced.
    paradat.doclist = doclist  ##  full corpus
    ##  first deal with mixed outcome types :
    paradat.catlist = sorted(list(set(outcomes)))
    paradat.catvals = set(paradat.catlist)
    paradat.cats = paradat.seencats = len(paradat.catvals)
    if paradat.cats < 65:
        print("Text types :",paradat.catvals)
    else:
        print("No. of text types =",paradat.cats)

    ##  parashow(paradat)

    ##  indirectly import module containing Docalib class :
    Catsoft = loadlib(paradat.libname)  ##  no nice error-message if it fails
    print("Text-classifier s/w successfully loaded from library :")
    print(Catsoft)
    print("[Expected to contain definition of class Docadat]")
    ##  if this doesn't work, might as well fail early.
    modinfo = Catsoft.Docadat(paradat,doclist)  ##  establish library
    ##  obtain extra docalib-specific parameters from paraline, if any :
    modinfo.loadpars(paradat)
    
    paradat.docs = len(doclist)
    assert paradat.docs == len(outcomes)
    print("Number of texts =",len(doclist))
    print("Number of tokens=",paradat.tottoks)
    print("Longest =",paradat.maxsize,"tokens.")
    paradat.meansize = sum(paradat.sizelist) / paradat.docs
    print("Mean size =",round(paradat.meansize,2))
    paradat.midsize = midvalue(paradat.sizelist)
    print("Median size =",paradat.midsize)
    paradat.minsize = min(paradat.sizelist)
    print("Smallest =",paradat.minsize)

    ##  what if counted != fnum ?
    if paradat.docs < Mindocs:
        print("Not enough data!")
        print("Program halts.")
        return  -2

    ##  prepare basic category info :
    catinfo = {}
    for c in paradat.catvals:
        catinfo[c] = Sack(c)
        catinfo[c].tottoks = 0
        catinfo[c].docs = 0
    ##  cumulate sizes by category :
    for d in doclist:
        catinfo[d.outcome].tottoks += d.size
        catinfo[d.outcome].docs += 1
    ##  testing :
    tmax = 0
    for c in paradat.catlist:
        if paradat.cats < 65:
            print(c,catinfo[c].docs,catinfo[c].tottoks)
        else:
            print(paradat.cats,"categories is too many!!")
            print("Program halts.") ; return  -3
        if catinfo[c].docs < 2:
            print("category",c,"has only 1 example.\nProgram halts!")
            return  -4
        if catinfo[c].docs < 3:
            print("category",c,"has only 2 examples.\nProgram halts!")
            return  -5
        if catinfo[c].tottoks > tmax:
            tmax = catinfo[c].tottoks
            paradat.refcode = c  ##  reference class is largest
        ##  also precompute relative sizes :
        catinfo[c].propsize = catinfo[c].tottoks / (paradat.tottoks+Tiny)
    print("reference category : ",paradat.refcode)

    ##  make single-string version of each text (move into glib.getdocs() ?) :
    paradat.totchars = 0
    for d in doclist:
        d.text = ' ' + ' '.join(d.toklist) + ' '
        ##  n.b. if ChinJaKo, better without blanks ?
        d.textlen = len(d.text)
        paradat.totchars += d.textlen
        if paradat.cats < 65:  ##  prepare for classification
            d.classnum = paradat.catlist.index(d.outcome)
            ##  also attach category number as well.
        ##  also make token frequency tables :
        d.freqtab = getfrex(d.toklist)
        ##  d.freqtab = Catsoft.getfrex(d.toklist)

    print("total characters =",paradat.totchars)
    for j in range(0):  ##  sanity check
        d = doclist[j]
        print(d.dnum,d.name,d.outcome,d.size,d.textlen)
        print(d.text[:69])

    ##  prepare o/p files :
    if not os.path.isdir(paradat.outpath):
        os.makedirs(paradat.outpath)
    if paradat.outfile:
        logfile = paradat.outfile
    else:
        logfile = paradat.outpath + paradat.jobname + Logtail + ".txt"
    paradat.fo = open(logfile,'w',encoding="utf8")
    paradat.logfile = paradat.fo  ##  synonym
    if os.path.isfile(paradat.dumpfile):
        dumpname = paradat.dumpfile
    else:
        dumpname = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    paradat.dumpfile = open(dumpname,'w',encoding="utf8")
    ##  also need separate results file :
    if os.path.isfile(paradat.listfile):
        listfile = paradat.listfile
    else:
        listfile = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listfile,'w',encoding="utf8")
    ##  also need file for o/p of catmodel info :
    if os.path.isfile(paradat.modsfile):
        modsfile = paradat.modsfile
    else:
        modsfile = paradat.outpath + paradat.jobname + Modtail + ".txt"
    paradat.modsfile = open(modsfile,'w',encoding="utf8")
    
    if paradat.fulldump:
        paradump(paradat,paradat.fo)
    else:
        parashow(paradat,paradat.fo)
        
    showobj(modinfo,paradat.fo)
    if hasattr(modinfo,"modprep"):
        modinfo.modprep(paradat)  ##  once-only preparation, e.g. snippets
    
    ##  main testing loop :
    testnum = Mintries
    testsize = int(paradat.docs**0.5)  ##  size of test subsample
    print(testnum)
    ncycles = 0
    docnums = list(range(paradat.docs))  ##  all indices
    notused = set(docnums)  ##  indices of never-used cases
    results = []
    while len(results) < testnum:
        ##  select training & test subsets :
        if len(notused) > testsize:  ##  pick from underused items
            testsubs = random.sample(notused,testsize)
        elif len(notused) == testsize:
            testsubs = sorted(notused)
        elif len(notused) < 1:  ##  pick from full list
            testsubs = random.sample(docnums,testsize)  ##  test sample indices
        else:  ##  awkward number of unused indices
            testsubs = list(notused)
            ##  if len(testsubs) < testsize:  ##  redundant belt&braces ?
                ##  tx = random.sample(docnums,testsize-len(testsubs))
                ##  testsubs += tx
            while len(testsubs) < testsize:
                tx = random.choice(docnums)
                if not (tx in testsubs):
                    testsubs.append(tx)
        assert len(testsubs) == testsize
        ##  training sample is those absent from testsubs.
        paradat.testsubs = testsubs
        notused = notused - set(testsubs)
        ##  pick training & test sets :
        testdox = []
        traindox = []
        for j in range(paradat.docs):
            if j in testsubs:
                testdox.append(doclist[j])
            else:
                traindox.append(doclist[j])
        ##  make category models :
        modinfo.makemods(paradat,traindox)
        ##  print(modinfo.vocsize)
        ##  print(traindox[0].trimtoks[:800])
        ##  test ruleset on test cases :
        testrex = ruletest(testdox,paradat,modinfo,"testmode")
        ##  record results :
        results.extend(testrex)
        ncycles += 1
        print(ncycles,len(results))
        if ncycles > testnum:
            break

    if len(notused) <= 20:
        print("unused test cases :",sorted(notused))
    else:
        print("unused test cases =",len(notused))
    print(len(results),"trials.")
    ##  summarize results :
    hits = 0
    rc = 0
    paradat.frontage = 0
    if results:
        frontage = len(results[0]) - paradat.seencats
        ##  number of items b4 actual catvals (pred&true cats just b4 that)
        paradat.frontage = frontage
        sortlist = outprep(paradat,results,paradat.docs)
        ##  reslist(paradat,sortlist)
        headlist = ["mode","ok","textnum","filename","predcat","truecat"]
        headline = Htab.join(headlist+[str(cn) for cn in paradat.catlist])
        print(headline,file=paradat.dumpfile)
    results1 = []  ##  save a copy as original gets altered.
    for r in results:
        rc += 1
        results1.append(r[:])
        if r[frontage-2] == r[frontage-1]:  ##  fragile !!
            hits += 1
        ##  else:
            ##  print(rc,r)
        outline = vecline(r,7)
        print(outline,file=paradat.dumpfile)    

    ##  display subsampling summary :
    ##  print(hits,round(100.0*hits/rc,4))
    showres(results,paradat,"testmode")
    ##  parashow(paradat,paradat.listfile)  ##  o/p header
    toplines(paradat,["progname","dateline","libname","paraline",
                      "trainmet","testmeta","targvar"])
    if results:
        print("\n==== Subsampling trial :",file=paradat.listfile)
        reslist(paradat,sortlist,paradat.listfile)
    showres(results,paradat,"testmode",paradat.listfile)

    ##  create rule from full corpus :
    modinfo.makemods(paradat,doclist)
    
    ##  show model info :
    modinfo.showmods(paradat,paradat.modsfile)

    ##  apply rule to holdout sample if given :
    testrecs = []
    if paradat.testmeta:
        if not os.path.isfile(paradat.testmeta):  ##  too severe ??
            raise RuntimeError("invalid test metafile given : " + paradat.testmeta + '!')
        ##  copy parameter data :
        paradat2 = paracopy(paradat,paradat.fo)
        print()
        metaname = getattr(paradat2,"testmeta",'')
        paradat2.outcomes = metagrab(metaname,paradat2)
        testdocs = getdocs(paradat2,paradat2.filelist,paradat2.outcomes,Maxtoks)
        ##  size limit enforced.
        paradat2.docs = len(testdocs)
        ##  save string versions of toklist :
        paradat2.totchars = 0
        for d in testdocs:
            d.text = ' ' + ' '.join(d.toklist) + ' '
            d.textlen = len(d.text)
            paradat2.totchars += d.textlen
            d.freqtab = getfrex(d.toklist)
        print("total characters =",paradat2.totchars)
        ##  showobj(paradat2)
        assert paradat.catlist == paradat2.catlist  ##  else trouble !!
        ##  test ruleset on test cases :
        testmode = "holdout"
        ##  here need to gather stats for zedscore calx?
        testrecs = ruletest(testdocs,paradat,modinfo,testmode)
        showres(testrecs,paradat2,testmode)
        print("\n==== Holdout trial :",file=paradat.listfile)
        ##  print(metaname+NL,file=paradat.listfile)
        ##  showres(testrecs,paradat2,testmode,paradat.listfile)
        hits = 0
        rc = 0
        for r in testrecs:
            rc += 1
            if r[paradat.frontage-2] == r[paradat.frontage-1]:  ##  must be just b4 simvals
                hits += 1
            outline = vecline(r,7)
            print(outline,file=paradat.dumpfile)
        ##  print(hits,rc)

    ##  final phase :
    if testrecs:  ##   no point unless some holdout cases done
        ##  collate stats from known-answer phase :
        vecdict = postvecs(paradat,results)
        for catname in paradat.catlist:
            vecdict[catname].sort()
            ##  xvec = vecdict[catname]
            ##  bs = bastats(xvec)
            ##  md = midvalue(xvec)
            ##  print(catname,md,bs.mean,bs.std2)
            ##  print(len(xvec),xvec[:7])
        ##  print(os.path.splitext(paradat.dumpfile.name))
        postname = os.path.splitext(paradat.dumpfile.name)[0]
        paradat.postname = postname + Posttail + ".dat"
        antepost = os.path.exists(paradat.postname)
        if paradat.postcode < 1:  ##  overwrite
            paradat.postdump = open(paradat.postname,'w',encoding="utf-8")
        else:  ##  append
            paradat.postdump = open(paradat.postname,'a',encoding="utf-8")
        ##  posthoc phase :
        sortlist = outprep(paradat,testrecs,len(testrecs))
        reslist(paradat,sortlist,paradat.listfile)
        showres(testrecs,paradat2,testmode,paradat.listfile)
        postlist = posthoc(paradat,testrecs,vecdict)
        ##  dump data :
        hits = rc = 0
        for p in postlist:
            r = p[-1]  ##  last item is data record
            rc += 1
            if r[paradat.frontage-2] == r[paradat.frontage-1]:  ##  must be just b4 simvals
                hits += 1
            outline = vecline(r,7)
            print(outline,file=paradat.dumpfile)
        ##  show on list file :
        print("\n==== Posthoc ranking :",file=paradat.listfile)
        ##  reslist(paradat,postlist,paradat.listfile)
        ##  also identify inconsistencies :
        ##  compare(paradat,sortlist,postlist,paradat.listfile)
        ##  P.S. :
        ##  for catname in paradat.catlist:
            ##  cx = vecdict[catname]
            ##  print(len(cx),catname,cx[:2],cx[-7:])
        lastout = postage(paradat,results1,sortlist,vecdict)
        ##  print(NL,file=paradat.listfile)
        ##  reslist(paradat,lastout,paradat.listfile)
        conflist(paradat,lastout,paradat.listfile)
        h = "credence\tok\ttextnum\tfilename\tpredcat\ttruecat\tconfidence\tcongruity"
        h += "\tlibname"
        ##  print(paradat.postcode,paradat.postname,antepost)
        ##  print(os.path.exists(paradat.postname))
        ##  print(os.path.isfile(paradat.postname))
        if paradat.postcode < 1 or not antepost:
            print(h,file=paradat.postdump)  ##  needs headline
        ##  only write header if not appending to existing posthoc file.
        nickname = paradat.libname
        if nickname.startswith("docalib_"):
            bits = nickname.split('_')
            
            if len(nickname) > 8:
                nickname = bits[1]
        for x in lastout:
            r = x[-1]
            ##  print(r[:paradat.frontage])
            lout = r[:paradat.frontage] + [x[-3],x[-2]]  ##  + conf, congruity
            outline = vecline(lout,7)
            outline = outline + Htab + nickname
            print(outline,file=paradat.postdump)
        paradat.postdump.close()

    ##  finale :
    print("Main output listed on :",os.path.abspath(paradat.listfile.name))
    print("Parameter settings on :",os.path.abspath(paradat.logfile.name))
    print("Export info dumped on :",os.path.abspath(paradat.dumpfile.name))
    ##  print(os.path.getsize(paradat.dumpfile.name))
    paradat.dumpfile.close()
    paradat.listfile.close()
    paradat.logfile.close()
    paradat.modsfile.close()
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff

    ui = pybyebye()  ##  fine as long as no errors
    return  0




##  WINDIST :
def windist (svec):
    "returns adjusted margin of victory in score vector."

    ns = len(svec)
    if ns < 2:
        return  0.0
    topscore = max(svec)
    if ns <= 2:
        return  topscore - min(svec)
    ##  3 or more scores :
    rest = svec[:]
    rest.remove(topscore)
    hi = max(rest)  ##  next highest
    lo = min(rest)
    w = [1] * (ns-1)  ## weight vector
    prodsum = 0.0  ##  to give weighted mean
    for j in range(ns-1):
        x = rest[j]
        if x == lo:
            w[j] = 1.0 / Gold
        elif x == hi:
            w[j] = Gold
        prodsum += w[j] * x

    return  topscore - prodsum / sum(w)


##  POSTAGE :
def postage (paradat,results,sortlist,vecdict):
    "another attempt at posthoc confidence ratings."

    ##  print(NL,"frontage =",paradat.frontage)
    nres = len(results)
    goodvals = []
    badvals = []
    ##  review results :
    for j,r in enumerate(results):
        tc = r[paradat.frontage-1]
        tv = max(r[paradat.frontage:])
        tote = sum(r[paradat.frontage:])
        sbar = (tote-tv) / (paradat.cats-1)
        difscore = tv-sbar
        if j < 0:
            print(j,r)
            print(tc,tv,difscore)
        if r[1] == Plusmark:  ##  correct answer
            goodvals.append(difscore)
        else:  ##  mistake
            badvals.append(difscore)
    goodvals.sort()
    goodones = len(goodvals)
    badvals.sort()
    badones = len(badvals)
    print(goodones,badones)
    ##  print("sortlist")
    outlist = []
    for j,r in enumerate(sortlist):
        difscore = r[0]
        ##  print(j,r)
        rx = r[2]
        ##  print(rx)
        truecat = rx[paradat.frontage-1]
        predcat = rx[paradat.frontage-2]
        topscore = max(rx[paradat.frontage:])
        gw = binwins(difscore,goodvals)
        bw = binwins(difscore,badvals)
        ##  gx = [(difscore>x) for x in goodvals]
        ##  gx = sum(gx)
        ##  c1 = (gw + 0.5) / (gw+(badones-bw)+1)
        c1 = (gw + 0.5*bw + 0.5) / (goodones + badones + 1.0)
        ownwins = binwins(topscore,vecdict[predcat])
        n = len(vecdict[predcat])
        c2 = (ownwins+0.5) / (n+1.0)
        ##  print(predcat,topscore,difscore,gw,gx,badones-bw,c1)
        ##  print(ownwins,n,ownwins/n)
        cred = math.sqrt(c1*c2)  ##  geomean
        ##  print(cred)
        outlist.append([cred,(sum(rx[paradat.frontage:])-topscore),c1,c2,rx])

    return  sorted(outlist,reverse=True)


##  CONFLIST :
def conflist (paradat,sortlist,fo=sys.stdout):
    "displays sorted results with confidence ratings."

    fo.write(NL)
    h = "rank credence  filename" + (' ' * Longfile) + "pred:true       confidence  congruity"
    print(h,file=fo)
    loutform = "{0:4d} {1:8.4f}  {2:" + str(Longfile) + "s}"
    marklist = []
    for j,s in enumerate(sortlist):
        sortval = s[0]  ##  sort value is first
        r = s[-1]  ##  data line is last
        c2 = s[-2]  ##  compat
        c1 = s[-3]  ##  confidence
        r[0] = sortval
        mark = r[1]
        marklist.append(mark)
        svec = r[paradat.frontage:]  ##  model scores
        ##  print(j+1,r[:paradat.frontage])
        lout = loutform.format(j+1,sortval,r[3][:Longfile])
        predcat = str(r[paradat.frontage-2])[:9]
        truecat = str(r[paradat.frontage-1])[:9]
        lout = lout + "  " + predcat.rjust(9) + ' ' + mark + ' ' + truecat.ljust(9)
        vout = " {0:9.4f} {1:8.4f}".format(c1,c2)
        print(lout+vout,file=fo)

    marx = ''.join(marklist)
    print(marx,file=fo)
    
    return

            
##  OUTPREP :
def outprep (paradat,results,listsize):
    "prepares results for listing."

    sortlist = []
    seen = set()
    for j,r in enumerate(results):
        svec = r[paradat.frontage:]
        smax = max(svec)
        sbar = (sum(svec)-smax) / (len(svec)-1)  ##  mean of rest
        ix = r[2]
        if not (ix in seen):  ##  only show first repetition
            seen.add(ix)
            sortval = smax - sbar
            sortlist.append([sortval,-min(svec),r])
        if len(sortlist) >= listsize:
            break

    sortlist.sort(reverse=True)
    
    return  sortlist


##  RESLIST :
def reslist (paradat,sortlist,fo=sys.stdout):
    "displays sorted results from sortlist."

    fo.write(NL)
    h = "rank   weight  filename" + (' ' * Longfile) + "pred:true         predval  meanrest"
    print(h,file=fo)
    loutform = "{0:4d} {1:8.4f}  {2:" + str(Longfile) + "s}"
    marklist = []
    for j,s in enumerate(sortlist):
        sortval = s[0]  ##  sort value is first
        r = s[-1]  ##  data line is last
        r[0] = sortval
        mark = r[1]
        marklist.append(mark)
        svec = r[paradat.frontage:]  ##  model scores
        smax = max(svec)
        sbar = (sum(svec)-smax) / (len(svec)-1)  ##  mean of rest
        ##  print(j+1,r[:paradat.frontage])
        lout = loutform.format(j+1,sortval,r[3][:Longfile])
        predcat = str(r[paradat.frontage-2])[:9]
        truecat = str(r[paradat.frontage-1])[:9]
        lout = lout + "  " + predcat.rjust(9) + ' ' + mark + ' ' + truecat.ljust(9)
        vout = " {0:9.4f} {1:8.4f}".format(smax,sbar)
        print(lout+vout,file=fo)

    marx = ''.join(marklist)
    print(marx,file=fo)
    
    return


##  POSTHOC :
def posthoc (paradat,testrecs,vecdict):
    "post-test using stats from first phase."

    ##  won't be called if testrecs is empty.
    catstuff = {}
    for catname in paradat.catlist:
        catstuff[catname] = Sack(catname)
        catstuff[catname].xvec = vecdict[catname]  ##  already in ascending order
        ##  bs = bastats(vecdict[catname])
        catstuff[catname].nx = len(vecdict[catname])
        ##  catstuff[catname].mean = bs.mean
        ##  catstuff[catname].sd = bs.std2
        md = midvalue(catstuff[catname].xvec)
        catstuff[catname].md = md
        ##  headpars(catstuff[catname])

    hits = 0 ; conf = 0
    posthits = 0 ; seencats = 0
    sortlist = []
    holdouts = len(testrecs)
    for tr in testrecs:
        svec = tr[paradat.frontage:]  ##  score vector
        truecat = tr[paradat.frontage-1]
        predcat = tr[paradat.frontage-2]
        ##  print(tr)
        textname = tr[3]
        ##  print(truecat,textname)
        if truecat in paradat.catlist:
            seencats += 1
        wsum = 0.0 ; ww = 0.0
        wvec = [0.0] * paradat.cats
        for c,catname in enumerate(paradat.catlist):
            s = svec[c]
            w = binwins(s,catstuff[catname].xvec)
            wr = (w+0.5) / (1.0+catstuff[catname].nx)
            wsum += wr
            wvec[c] = wr
            ##  z = (s - catstuff[catname].mean) / catstuff[catname].sd
            ##  zedscores too wild !!
            if catname == predcat:  ##  this class was predicted
                ww = wr
                if wr >= 0.5 and (truecat in paradat.catlist):
                    conf += 1  ##  confident
                    if catname == truecat:
                        hits += 1
        ##  sortval = (ww-(wsum-ww))
        ##  sortlist.append([sortval,textname,tr])
        ##  choice on revised basis :
        wmax = max(wvec)
        wpos = wvec.index(wmax)
        topvals = wvec.count(wmax)
        predcat = paradat.catlist[wpos]
        if predcat == truecat:
            posthits += 1.0 / topvals
            mark = Plusmark
        elif truecat in paradat.catlist:
            mark = Negmark
        else:
            mark = Oddmark  ##  couldn't be right
        sortval = wmax - sum(wvec) / paradat.cats  ##  wsum
        ##  wvec.remove(ww)
        ##  w2 = max(wvec)
        ##  sortval = ww - sum(wvec) / (paradat.cats-1)
        ##  sortval = ww - w2
        ##  sortval = ww / wsum
        ##  sortval = max(svec) - sum(svec)/len(svec)
        tx = tr[:paradat.frontage] + wvec
        tx[0] = "posthoc"
        tx[1] = mark
        tx[paradat.frontage-2] = predcat
        sortlist.append([sortval,-min(wvec),tx])
        ##  print(wmax,ww,w2,wsum,round(sortval,4))

    ##  if seencats > 0:
        ##  print("posthits",posthits,seencats,round(100*posthits/seencats,4))
    sortlist.sort(reverse=True)
        
    return  sortlist


##  BINWINS :
def binwins (x,xvec):
    "number of items that x exceeds in xvec, presumed sorted upwards."

    n = len(xvec)
    if n < 1:
        return  0
    lo = 0 ; hi = n-1
    ##  binsearch to approx location of x value :
    while hi >= lo:
        mp = (lo+hi) // 2
        if xvec[mp] < x:
            lo = mp + 1
        elif xvec[mp] > x:
            hi = mp - 1
        else:  ## eq
            while mp >= 0 and xvec[mp] == x:
                mp -= 1  ##  down below equalities
            lo = mp
            break
    mp = min(lo,hi)  ##  definitely not too high
    mp = max(0,mp)
    ##  print(lo,mp,hi)
    w = mp ; eq = 0
    ##  step thru till beaten
    for j in range(mp,n):
        if x > xvec[j]:
            w += 1
        elif x == xvec[j]:
            eq += 1
        else:
            break

    return  w + 0.5 * eq


##  POSTVECS
def postvecs (paradat,results):
    "extracts scores of correct hindcasts, by category, from results list."

    vecdict = {}
    for r in results:
        tc = r[paradat.frontage-1]  ##  true catname
        tk = paradat.catlist.index(tc)  ##  offset of true-category score
        ##  print(frontage,r)
        cs = r[paradat.frontage+tk]
        ##  print(tc,tk,cs)
        if tc in vecdict:
            vecdict[tc].append(cs)
        else:
            vecdict[tc] = [cs]  ##  first time
        
    return  vecdict


##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  LOADLIB :
def loadlib (libname):
    "tries to import a library named at run time, using local namespace."

    locdict = {"grabbed" : '', "libname" : libname}
    impline = "import " + libname + " as grabbed"
    exec(impline,globals(),locdict)
    ##  print(':',locdict["grabbed"])

    return  locdict["grabbed"]
    ##  should avoid altering global namespace.


##  SPHERICS :
def spherics (vals,tc):
    "computes 'spherical score' of probability estimates."

    vtot = sum(vals)
    if vtot <= 0.0:
        return  0.0  ##  dodgy
    if vtot == 1.0:
        pvec = vals
    else:  ##  make into probs
        pvec = [v/vtot for v in vals]
    ps = [p**2 for p in pvec]
    ps = sum(ps)

    return  pvec[tc] / (ps ** 0.5)
    ##  trouble if tc (true category position) isn't in range!


##  ASPROBS :
def asprobs (xvec,epsilon=1e-12):
    "forces xvec (somewhat crudely) 2b a list of quasi-probabilities."

    nx = len(xvec)
    if nx < 1:
        return  []
    xmax = max(xvec) ; xmin = min(xvec)
    xtot = sum(xvec)  ##  will fail if nonnumeric
    if xmax <= 1.0 and xmin >= 0.0 and abs(1.0-xtot) < epsilon:
        ##  probably probabilities already
        return  xvec[:]
    ##  similarities don't look like probabilities :
    if xmax - xmin < epsilon:
        return [1.0/nx] * nx  ##  uninformative
    if xmax > 0 and xmin >= 0:  ##  just scale to total 1
        pvec = [x/xtot for x in xvec]
        return  pvec
    ##  kludge :
    pvec = [(x-xmin)+epsilon for x in xvec]
    ptot = sum(pvec)
    return [p/ptot for p in pvec]
    ##  expects similarity scores, so with distances invert before calling.


##  SHOWRES :
def showres (results,paradat,testmode,fo=sys.stdout):
    "shows results of testing process."

    if len(results) < 1:
        return
    ##  compute indicators :
    cmat = {}  ##  confusion matrix
    hits = 0
    tent = 0.0  ##  total log penalty
    rc = 0
    seen = 0
    stot = 0.0
    minival = Huge
    ##  vectors for noninteger evaluation scores :
    diflist = []
    binlist = []
    maxvals = [] ; meanvals = []
    truevals = []
    probword = ''
    frontage = len(results[0]) - paradat.seencats
    ##  frontage gives # of items b4 catsims.
    ppos = frontage - 2  ##  position of predcat
    tpos = frontage - 1  ##  position of truecat
    ##  print("frontage",frontage)
    ##  print(paradat.seencats,paradat.cats)
    ##  print(ppos,tpos,results[0])
    ##  print(paradat.catlist)
    for r in results:
        rc += 1
        pair = (r[ppos],r[tpos])  ##  predcat, truecat must be just b4 catvals
        ##  print(pair,type(pair[0]),type(pair[1]))
        try:  ##  catlist copied over unchanged from first paradat
            tp = paradat.catlist.index(pair[1])  ##  true position
        except ValueError as xxxx:
            tp = -1  ##  catlabel never seen b4
        if tp >= 0:  ##  class seen in training sample
            predpos = paradat.catlist.index(pair[0])
            seen += 1
            catvals = r[frontage:]  ##  simvals come after extra items
            if abs(sum(catvals)-1.0) > 1e-8:
                probword = "pseudo-"
                ##  treat as pseudo-probabilities.
            pvec = asprobs(catvals)  ##  try2 turn vals into probs
            assert catvals[predpos] == max(catvals)  ##  just checking
            ss = spherics(pvec,tp)
            stot += ss
            ##  print(catvals,tp,file=fo)
            ##  print(pvec,ss,file=fo)
            ##  if pair[0]!=pair[1]:
                ##  print('?',file=fo)
            ##  also compute entropy :
            pr = pvec[tp]
            e = math.log(max(1e-10,pr),2.0)  ##  kludge tiny probs
            ##  if e < -Gold:  ##  show large misses
                ##  print(catvals,sum(catvals),file=fo)
                ##  print(pvec,sum(pvec),'@',tp,e,file=fo)
            tent -= e
            maxvals.append(max(catvals))
            minival = min(minival,min(catvals))
            truevals.append(catvals[tp])
            ##  also compute excess over mean of rest (for pb corr of deltas) :
            meanval = sum(catvals) / len(catvals)
            meanvals.append(meanval)
            meanrest = (sum(catvals)-catvals[predpos]) / (len(catvals)-1)
            ##  won't run with less than 2 cats, surely ??
            diff = catvals[predpos] - meanrest
            diflist.append(diff)
            if pair[0]==pair[1]:
                binlist.append(1)
            else:
                binlist.append(0)
            assert len(catvals) == paradat.seencats
            ##  difs = [c-meanval for c in catvals]
            ##  bins = [0] * len(catvals)
            ##  bins[tp] = 1
            ##  binlist.extend(bins)  ##  1 if true category, else zero
            ##  diflist.extend(difs)  ##  difs from expected value
        if pair[0]==pair[1]:
            hits += 1
        if pair in cmat:
            cmat[pair] += 1
        else:
            cmat[pair] = 1

    ##  show confusion matrix :
    ptab,rtab = precall(cmat)
    keylist = sorted(cmat.keys())  ##  list of keypairs
    wide = max([len(str(k)) for k in keylist])
    wide = wide // 2
    ##  matshow & precall expect pred,true order :
    matshow(fo,cmat)
    sk = sckappa2(cmat,paradat.catlist)  ##  catlist better than keylist items
    k2 = round(sk[0],4)
    ##  print(hits,sk[2])
    assert hits == sk[2]
    print("Kappa value = ",k2,file=fo)
    print("Precision (%) by category :",file=fo)
    for c in sorted(ptab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(ptab[c],4),file=fo)
    print("Recall (%) by category :",file=fo)
    for c in sorted(rtab.keys()):
        print(str(c)[:Longname].ljust(wide),
              round(rtab[c],4),file=fo)

    ##  display indicators :
    print("\ncases =",rc,file=fo)
    print("cases with unseen category labels =",rc-seen,file=fo)
    if seen > 0:
        if seen == 1:
            word = "case"
        else:
            word = "cases"
        if seen < rc:
            print("cases with known category labels =",seen,file=fo)
            print("[results below, till '*', only apply to these",seen,word+']',
                  file=fo)
        print("hits = ",hits,file=fo)
        percent = 100 * hits / seen
        print("percent hits =",round(percent,2),file=fo)
        prie = priemat(paradat.catlist,cmat)
        print("proportional reduction in error =",round(prie,4),file=fo)
        ##  probword blank if all vals in 0..1, else 'pseudo-" :
        ae = tent / seen
        xe = -math.log(1/len(catvals),2)
        ##  print(ae,xe)
        print("relative mean "+probword+"entropy gain =",round(1.0-ae/xe,4),
              file=fo)
        ##  treat as probabilities :
        ss = stot / seen
        print("mean "+probword+"spherical score = ",round(ss,4),file=fo)
        mg = margain(diflist,binlist)
        print("marginal gain =",round(mg,4),file=fo)
        gs = gapscore(maxvals,meanvals,truevals)
        print("gapscore =",round(gs,4),file=fo)
        if seen < rc:
            print('*',file=fo)
        
    return


##  MARGAIN :
def margain (diflist,binlist):
    "marginal gain."

    n = len(diflist)
    if n < 2:  return  0.0
    assert n == len(binlist)
    goodgaps = 0 ; allgaps = 0.5 / n
    for j in range(n):
        v = diflist[j]  ##  margin: max(vals) - mean of other vals
        allgaps += v
        if binlist[j] == 1:  ##  correct decision
            goodgaps += v

    return  goodgaps / allgaps
    

##  GAPSCORE :
def gapscore (maxvals,meanvals,truevals):
    "parametric analogy to proportional reduction in error."

    n = len(maxvals)
    assert n == len(meanvals) and n == len(truevals)
    meansum = sum(meanvals)
    ##  print(sum(maxvals),meansum,sum(truevals))
    gs = sum(truevals) - meansum
    dnom = sum(maxvals) - meansum  ##  exceeds zero unless all equals

    return  gs / max(1e-16,dnom)  ##  should be between 0 & 1.


##  PRIEMAT :
def priemat (catlist,cmat):
    "proportional reduction in error on confusion matrix held as pairtab."

    cats = len(catlist)
    if cats < 2:
        return  0
    ##  print(cats)
    ##  compute marginal totals :
    predtots = {} ; truetots = {}
    correct = 0 ; grandtot = 0
    tmax = 0
    for c1 in catlist:  ##  only known cats used
        for c2 in catlist:
            pair = (c1,c2)
            f = cmat.get(pair,0)
            grandtot += f
            if c1 == c2:  correct += f
            if c1 in predtots:  ##  row is predicted
                predtots[c1] += f
            else:
                predtots[c1] = f
            t = truetots.get(c2,0)  ##  column is actual
            truetots[c2] = t + f
    for t in truetots:
        if truetots[t] > tmax:
            tmax = truetots[t]
    ##  compute base rates :
    baserate = 1.0 / cats
    maxrate = tmax / grandtot
    bushrate = 0 ; kaprate = 0
    for c in catlist:
        pr = predtots.get(c,0) / max(1,grandtot)
        tr = truetots.get(c,0) / max(1,grandtot)
        bushrate += tr*tr
        kaprate += pr*tr
    ##  print(baserate,bushrate,kaprate,maxrate)
    ##  p3 = (bushrate+kaprate+maxrate) / 3.0
    ##  print(p3)
    pc = (bushrate+maxrate) * 0.5  ##  chance expected proportion
    ##  compute prie :

    prie = (correct/max(1,grandtot) - pc) / (1.0 - pc)

    return  prie


##  TOPLINES :
def toplines (paradat,parnames,outfile=''):
    "shows selected parameter values as o/p header."

    if not outfile:
        if paradat.listfile:
            outfile = paradat.listfile
        else:
            outfile = sys.stdout
    outpars = Sack(paradat.id)
    for p in parnames:
        if hasattr(paradat,p):
            setattr(outpars,p,getattr(paradat,p,''))

    showobj(outpars,outfile)

    return


##  PRECALL :
def precall (cmat):
    "returns precisions & recalls of confusion matrix held as freqtab of pairs."

    c1 = set() ; c2 = set()
    pairlist = sorted(cmat.keys())  ##  order in pair: predcat,truecat
    for k1,k2 in pairlist:
        c1.add(k1) ; c2.add(k2)
    c1 = sorted(c1) ; c2 = sorted(c2)
    ptab = {} ; rtab = {}
    ##  obtain totals :
    for p in pairlist:
        f = cmat[p]
        pn = ptab.get(p[0],0)
        ptab[p[0]] = pn + f
        rn = rtab.get(p[1],0)
        rtab[p[1]] = rn + f
    ##  print(ptab,rtab)
    ##  compute percentages :
    for c in c1:
        ok = cmat.get((c,c),0)
        ptab[c] = 100.0 * ok / ptab.get(c,1)  ##  precision in class c
    for c in c2:
        ok = cmat.get((c,c),0)
        rtab[c] = 100.0 * ok / rtab.get(c,1)  ##  recall of class c

    return  (ptab,rtab)


##  MATSHOW :
def matshow (fo,cmat,longest=Longname):
    "write confusion matrix (dict of pairs) to already open file."

    fo.write(NL + "Confusion matrix :" + NL + NL)
    ##  make local category list (could include unknown cats in holdout mode) :
    cats = set()
    for k1,k2 in cmat.keys():
        cats.add(k1) ; cats.add(k2)
    catlist = sorted(cats)
    ##  find widest class name :
    w = 4
    for c in catlist:
        wx = len(str(c))
        if wx > w:
            w = wx
    w = min(longest,w)
    lout = "Truecat =".ljust(10+w)
    for c in catlist:
        lout = lout + str(c)[:longest].rjust(w+1)
    fo.write(lout + NL)
    for pred in catlist:
        lout = "Predcat : " + str(pred).ljust(w)
        for truecat in catlist:
            pair = (pred,truecat)
            f = cmat.get(pair,0)
            lout = lout + " {0:{1}d}".format(f,w)
        fo.write(lout + NL)

    fo.write(NL)
    return


##  RULETEST :
def ruletest (testdocs,paradat,modinfo,testmode):
    "applies rules to docs & returns category scores (quasi-probs?)."

    testrecs = []
    correct = 0
    for d in testdocs:
        scorevec = modinfo.modsims(d,paradat)
        ##  if distance-based, should matches or ruletest convert ??
        smax = max(scorevec)  ##  maximizing at this stage (e.g. probs)
        cmax = scorevec.index(smax)
        predcat = paradat.catlist[cmax]
        truecat = d.outcome
        ##  print(d.dnum,d.name,predcat,truecat)
        if not (truecat in paradat.catlist):  ##  no chance 2b correct
            ok = Oddmark
        elif predcat == truecat:
            correct += 1
            ok = Plusmark
        else:
            ok = Negmark
        outlist = [testmode,ok,d.dnum,d.name,predcat,truecat] + scorevec
        testrecs.append(outlist)

    ##  print(len(testdocs),correct,round(correct*100.0/len(testdocs),4))

    return  testrecs
    ##  each line must have scores @end preceded by predcat & truecat.



            
##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
