"Python3 utility routines for toccata system, by R.S. Forsyth."

##  TOCCLIB :
##
##  Python3 general/statistical utilities :
##  Copyright 2012, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 04/02/2013
##  last revision : 30/01/2017


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
    This copyrighted program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)
    
	
import  math, os, string, sys
import  random

from globdefs import *



##  NUMERIZE :
def numerize(s):
    "Converts s to a numeric value if possible, else returns None."

    try:
        v = int(s)
    except ValueError:
        try:
            v = float(s)
        except ValueError:
            v = None  ## means s doesn't look like a number
    return v


##  BASTATS :
def bastats (numlist):
    "basic stats routine for a list of numbers."

    fn = float(len(numlist))
    sum1 = 0.0
    sum2 = 0.0
    for x in numlist:
        sum1 += x
        sum2 += x**2

    statsack = Sack(0)
    ##  defaults for dodgy inputs :
    statsack.mean = sum1
    statsack.var1 = statsack.var2 = sum2
    statsack.std1 = statsack.std2 = math.sqrt(abs(sum2))
    if fn > 0.0:
        statsack.mean = sum1 / fn
        statsack.var1 = sum2 / fn - statsack.mean**2
        ##  print "$$",statsack.var1
        statsack.std1 = math.sqrt(max(0.0,statsack.var1))
        if fn > 1.0:
            statsack.var2 = statsack.var1 * (fn/(fn-1.0))
            statsack.std2 = math.sqrt(max(0.0,statsack.var2))

    return statsack


##  GEOMEAN :
def geomean (xvec):
    "computes geometric mean of vector of postive values."

    n = len(xvec)
    if n < 1:
        return  0.0
    sumlogs = 0.0
    for v in xvec:
        if v < Tiny:
            return  0.0  ##  if any zeroes, result is zero
        sumlogs += math.log(v)
    return  math.exp(sumlogs/n)



##  CORSUMS :
def corsums (x,y):
    "produces sums required for correlation coefficient, in a Sack()."

    cs = Sack(0)
    n = len(x)
    n = min(n,len(y))  ##  chops extra if unequal
    if n < 1:
        print("Problem in corsums!!")
        return cs
    sx = sy = 0.0
    sxsq = sysq = 0.0
    sumprods = 0.0
    for i in range(n):
        sx += x[i] ; sy += y[i]
        sxsq += x[i]**2 ; sysq += y[i]**2
        sumprods += x[i] * y[i]

    cs.n = float(n)
    cs.sumprods = sumprods
    cs.sx = sx
    cs.sy = sy
    cs.sxsq = sxsq
    cs.sysq = sysq
    return  cs


##  CORRVAL :
def corrval (cs,epsilon=1e-23):
    "computes correlation coefficient from sums etc. in Sack()."

    cnum = cs.sumprods - (cs.sx*cs.sy) / cs.n
    tx = cs.sxsq - cs.sx**2 / cs.n
    ty = cs.sysq - cs.sy**2 / cs.n
    ##  try:
        ##  dnom = math.sqrt(tx*ty)
    ##  except ValueError as xxxx:
        ##  print(xxxx)
        ##  print(tx,ty,cnum,cs.n)
        ##  print(cs.sx,cs.sy,cs.sxsq,cs.sysq)
        ##  dnom = 0
    ##  dnom = math.sqrt(abs(tx*ty))  ##  danger of rounding errors
    dnom = math.sqrt(max(0.0,tx*ty))
    if dnom > epsilon:
        return  cnum / dnom
    elif tx < epsilon and ty < epsilon:
        return  1.0  ##  both flat
    else:
        return  0.0  ##  null case
    ##  semi-safe, but vulnerable to rounding errors.


##  RANX :
def ranx (xvec):
    "returns rank vector from raw data vector, with ties given mean positions."

    nx = len(xvec)
    ##  xvec.sort()  ##  ascending
    vtab = {}
    for x in xvec:
        fx = vtab.get(x,0)
        vtab[x] = fx + 1
    vvec = list(vtab.items())
    vvec.sort()
    ranktab = {}
    p = 0
    ranksum = 0
    for v in vvec:
        vx = v[0] ; vf = v[1]
        rx = (ranksum*2+vf)*0.5-0.5  ##  plus 0.5 if lowest rank is 1
        ranksum += vf
        ##  print(p,vx,vf,rx)
        ranktab[vx] = rx
        p += 1

    rankvec = []
    for x in xvec:
        rankvec.append(ranktab[x])
    return  rankvec  ##  returning ranktab more useful ?


##  MIDVALUE :
def midvalue (xvec):
    "simple median."

    vecsize = len(xvec)
    if vecsize < 1:
        return  0.0  ##  None better ?
    if vecsize % 2 == 1:  ##  uneven
        m = vecsize // 2
        return  sorted(xvec)[m]
    else:  ##  even
        m = (vecsize-1)//2
        v1,v2 = sorted(xvec)[m:m+2]
        return  (v1+v2)*0.5


##  HEDGESG :
def hedgesg (x1,x2):
    "effect-size measure for 2 independent groups: Cohen's d with pooled variance."

    n1 = len(x1) ; n2 = len(x2)
    if (n1+n2) < 4 or n1 < 1 or n2 < 1:
        return  0.0
    s1 = bastats(x1)
    s2 = bastats(x2)
    diff = s1.mean - s2.mean
    if (n1+n2) <= 4:  ##  puny sample
        return  diff / max(s1.std2,s2.std2)
    else:  ##  normal case
        pooledv = ((n1-1)*s1.var2 + (n2-1)*s2.var2) / (n1+n2-2)
        return  diff / math.sqrt(pooledv)


##  KAGDIST :
def kagdist (xvec,yvec,logbase=math.e):
    "error function used (on nonnegative numbers) in a Kaggle contest, 2013."

    ##  logbase makes a difference (though not to ranks).
    n = min(len(xvec),len(yvec))
    if n < 1:
        return  0.0  ##  suitable default ??
    dif2 = [(math.log(xvec[j]+1,logbase)-math.log(yvec[j]+1,logbase))**2 \
            for j in range(n)]

    return  math.sqrt(sum(dif2) / n)
    ##  can treat as distance function if both vecs scaled compatibly.


##  SCKAPPA2 :
def sckappa2 (ctab,codelist):
    "Siegel & Catellan's version of Kappa, for 2-way confusion matrix held as table."

    nc = len(codelist)
    chap = 0.0  ##  proportion agreement expected by chance
    totf = 0.0  ##  total frequency, all cells in table
    hits = 0
    codefrex = {}
    for k1 in codelist:
        ##  codelist holds all valid category codes.
        for k2 in codelist:
            pair = (k1, k2)
            f = ctab.get(pair,0)
            if k1 == k2:
                hits += f
            f1 = codefrex.get(k1,0)
            codefrex[k1] = f1 + f
            f2 = codefrex.get(k2,0)
            codefrex[k2] = f2 + f
            totf = totf + f
    totf = max(Tiny,totf)  ##  to prevent zero division
    for k in codelist:
        f = codefrex.get(k,0)
        pk = f / (2.0 * totf)
        chap += (pk * pk)
        ##  print k,f,pk,chap
    ##  chap is sum of squared probs.
    pa = hits / totf
    ##  print  hits, pa, totf
    k = (pa - chap) / max((1 - chap),Tiny)
    return  (k, chap,hits,pa,totf)


##  FOURFOLD :
def fourfold (xvec,yvec,xm=None,ym=None):
    "makes fourfold table frequencies with respect to threshold values."

    ##  if none given, medians used as thresholds :
    if xm == None:
        xm = midvalue(xvec)
    if ym == None:
        ym = midvalue(yvec)
    a = b = c = d = 0
    n = min(len(xvec),len(yvec))  ##  should be equal
    if n < 1:
        return  0.0
    for j in range(n):
        if xvec[j] > xm:
            if yvec[j] > ym:  ##  11
                a += 1
            else:  ##  10
                b += 1
        else:
            if yvec[j] > ym:
                c += 1  ##  01
            else:  ##  00
                d += 1

    return  (a,b,c,d)


##  TETRACOR :
def tetracor (a,b,c,d,slug=0.5):
    "tetrachoric correlation coefficient from fourfold table counts [11,10;01,00]."

    x = math.sqrt((a+slug)*(d+slug))  ##  attenuation factor added
    y = math.sqrt((b+slug)*(c+slug))
    n = x - y
    d = x + y + Tiny
    return  math.sin(0.5*math.pi*(n/d))


##  G2 :
def g2 (f1,n1,f2,n2):
    "special case of log-likelihood function for corpus frequency comparison."

    totfreq = f1 + f2
    grandtot = n1 + n2  ##  typically n1 is smaller, though not essential
    if totfreq < 4:
        return  (0.0,0,0)
    rate = totfreq / float(grandtot)
    e1 = rate * n1 ; e2 = rate * n2
    ##  print f1,n1,e1,"*",f2,n2,e2
    if e1 > Tiny and f1 > 0:
        x1 = f1 * math.log(f1/e1)
    else:
        x1 = 0.0
    if e2 > Tiny and f2 > 0:
        x2 = f2 * math.log(f2/e2)
    else:
        x2 = 0.0

    return (2*(x1+x2),x1,x2)  ##  returns a triple, item zero is g2
    ##  sign of x1 is directional indicator, x1 <= 0 indicates negative assoc.


##  SLUGGER :
def slugger (freqvec):
    "attenuates 1d frequency vector using quasi-bayes minimax heuristic (sqrt)."

    nc = len(freqvec)
    if nc < 2:
        return  freqvec  ##  can't work with single category
    tote = sum(freqvec)
    if tote <= 2:
        return  freqvec  ##  doesn't really make sense with tiny totals
    ##  better to return empty list as warning sign ??
    x = math.sqrt(tote)  ##  number of extra pseudo-cases
    ##  distribute pseudo-cases equally :
    slugged = [x/nc] * nc
    for c in range(nc):
        slugged[c] += freqvec[c]
    slugtot = sum(slugged)  ##  revised, increased total
    return  [fx/slugtot for fx in slugged]
    ##  returns pseudo-props,
    ##  which must be multiplied by suitable total to give pseudo-freqs.


def slugging (f1,n1,f2,n2):
    "attenuated rates."

    n = n1+n2
    if n <= 2:
        return  [0,0]  ##  no meaningful estimate
    m = (f1+f2) / n
    x1 = n1 ** 0.5 ; x2 = n2 ** 0.5
    r1 = (f1 + m*x1) / (n1 + x1)
    r2 = (f2 + m*x2) / (n2 + x2)

    return  [r1,r2]
    ##  should diminish contrast




if 0 > 0:
    for j in range(4):
        f1 = random.randint(1,55)
        n1 = random.randint(144,10000)
        f2 = random.randint(2,31)
        n2 = random.randint(233,20000)
        r1 = f1/n1 ; r2 = f2/n2
        print(f1,n1,f2,n2)
        print([r1,r2],r1-r2)
        sluggard = slugging(f1,n1,f2,n2)
        print(sluggard,sluggard[0]-sluggard[1])


if 0 > 0:  ##  moderation test :
    fvec = [14,6]
    fx = slugger(fvec)
    print(fvec)
    print([f / sum(fvec) for f in fvec])
    print(fx,sum(fx))
    bs = bastats(fx)
    print(bs.std1)
    zonk = [f * sum(fvec) for f in fx]
    print(zonk)  ##  revised frex
    gs = g2(fvec[0],12000,fvec[1],10000)
    gz = g2(zonk[0],12000,zonk[1],10000)
    print(gs[0],gz[0])
    fvec = [12000,10000]
    fx = slugger(fvec)
    print(fvec)
    print([f / sum(fvec) for f in fvec])
    print(fx,sum(fx))
    print([f * sum(fvec) for f in fx])
    
